/*
 * Copyright Staffan Gimåker 2006-2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef PEEKABOT_RENDERER_RELAY_HH_INCLUDED
#define PEEKABOT_RENDERER_RELAY_HH_INCLUDED


#include <boost/utility.hpp>

#include "Types.hh"
#include "Singleton.hh"


namespace peekabot
{
    class SceneObject;

    namespace renderer
    {
        class FrameCollector;

        namespace api
        {
            class Renderer;
            class EntityProxy;
        }
    }

    /**
     * \internal
     *
     * \brief Provides the glue needed to connect the main scene tree with the
     * renderer.
     *
     * Automagically takes care of making all objects in the scene tree
     * available and kept up to date in the renderer. It also provides methods to
     * render the scene, pick objects, etc.
     */
    class RendererRelay : public boost::noncopyable
    {
    public:
        RendererRelay();

        RendererRelay(const SceneObject *root);

        ~RendererRelay();

        /**
         * \brief Render a frame of the scene into the active widget.
         *
         * \remark This method is blocking and may take non-trivial time to
         * execute.
         *
         * \pre Must be called with a valid OpenGL context.
         *
         * \param camera_id ObjectID of the camera we should use to render the
         * scene.
         */
        void render(
            unsigned int w, unsigned int h, ObjectID camera_id,
            bool *visible_layers, renderer::FrameCollector *collector);

        /**
         * \brief Return the nearest object at the given coordinate, or 0 if
         * there is no object at that coordinate.
         *
         * \remark This method is blocking and may take non-trivial time to
         * execute.
         *
         * \pre Must be called with a valid OpenGL context.
         *
         * \param camera_id ObjectID of the camera we should use to render the
         * scene.
         */
        uint32_t pick(
            unsigned int w, unsigned int h, ObjectID camera_id,
            bool *visible_layers,
            int x, int y, float sensitivity);

        /**
         * \brief Initialize the renderer.
         *
         * \pre Must be called with a valid OpenGL context.
         */
        void initialize_renderer();

        /**
         * \brief Return true if there are pending updates to the renderer.
         */
        bool has_pending_renderer_updates() const throw();

        /**
         * \brief Process pending renderer updates.
         *
         * Pending updates are implcitly processed when pick() or render() are
         * called. This is method is useful to avoid excess memory usage when
         * render() nor pick() are called for extensive time periods.
         *
         * \remark This method is blocking and may take non-trivial time to
         * execute.
         *
         * \pre Must be called with a valid OpenGL context.
         */
        void process_renderer_updates();

        renderer::api::Renderer &get_renderer();

        const renderer::api::Renderer &get_renderer() const;

    private:
        class Impl;

        Impl *m_impl;
    };


    typedef Singleton<RendererRelay> TheRendererRelay;
}


#endif // PEEKABOT_RENDERER_RELAY_HH_INCLUDED
