/*
 * Copyright Staffan Gimåker 2006-2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cassert>
#include <cmath>
#include <boost/bind.hpp>

#include "RendererRelayImpl.hh"
#include "SceneObject.hh"

#include "CameraObject.hh"
#include "SceneObject.hh"
#include "GroupObject.hh"
#include "ScalableObject.hh"
#include "GeometricPrimitives.hh"
#include "Hinge.hh"
#include "Slider.hh"
#include "PointSetObject.hh"
#include "ModelObject.hh"
#include "GridObject.hh"
#include "PointCloud.hh"
#include "LineCloud.hh"
#include "OccupancyGrid2D.hh"
#include "Label.hh"

#include "renderer/api/Renderer.hh"
#include "renderer/api/CameraProxy.hh"
#include "renderer/api/SphereProxy.hh"
#include "renderer/api/CylinderProxy.hh"
#include "renderer/api/CubeProxy.hh"
#include "renderer/api/CircleProxy.hh"
#include "renderer/api/LineMeshProxy.hh"
#include "renderer/api/PointMeshProxy.hh"
#include "renderer/api/ModelProxy.hh"
#include "renderer/api/LabelProxy.hh"
#include "renderer/api/PolygonProxy.hh"
#include "renderer/api/OccupancyGrid2DProxy.hh"
#include "renderer/api/DummyProxy.hh"


using namespace peekabot;
using namespace peekabot::renderer::api;


namespace
{
    void set_grid_vertices_stub(
        LineMeshProxy *proxy,
        const GridObject *object)
    {
        std::vector<Vector3f> verts;
        unsigned int segments = object->get_segment_count();

        if( object->get_grid_type() == REGULAR_GRID )
        {
            float foo = segments/2.0f;

            for( unsigned int i = 0; i < segments+1; ++i )
            {
                verts.push_back(Vector3f(-foo+i, -foo, 0));
                verts.push_back(Vector3f(-foo+i,  foo, 0));
            }
            for( unsigned int i = 0; i < segments+1; ++i )
            {
                verts.push_back(Vector3f(-foo, -foo+i, 0));
                verts.push_back(Vector3f( foo, -foo+i, 0));
            }
        }
        else if( object->get_grid_type() == RADIAL_GRID )
        {
            float theta = object->get_central_angle();
            float offset = -theta/2;

            for( unsigned int i = 0; i < segments; ++i )
            {
                unsigned int r = i+1;
                unsigned int circle_segments = static_cast<unsigned int>(
                    ceilf((64+10*i) * theta/M_PI/2));

                for( unsigned int j = 0; j < circle_segments; ++j )
                {
                    verts.push_back(Vector3f(r*cosf(theta*j/circle_segments+offset),
                                             r*sinf(theta*j/circle_segments+offset), 0));
                    verts.push_back(Vector3f(r*cosf(theta*(j+1)/circle_segments+offset),
                                             r*sinf(theta*(j+1)/circle_segments+offset), 0));
                }
            }
        }
        else if( object->get_grid_type() == ANGULAR_GRID )
        {
            float theta = object->get_central_angle();
            float offset = -theta/2;

            unsigned int lines;
            if( fabsf(theta-2*M_PI) < 0.000001 )
                lines = segments;
            else
                lines = segments+1;

            for( unsigned int i = 0; i < lines; ++i )
            {
                verts.push_back(Vector3f(0,0,0));
                verts.push_back(Vector3f(cosf(theta*i/segments+offset),
                                         sinf(theta*i/segments+offset), 0));
            }
        }
        else
            assert( false );

        proxy->set_vertices(verts);
    }


    void set_transformation_stub1(
        ObjectEntityProxy *proxy,
        const SceneObject *object)
    {
        Matrix4f mtow = object->get_mtow();

        const ScalableObject *scalable(
            dynamic_cast<const ScalableObject *>(object));

        if( scalable )
        {
            mtow(0) *= scalable->get_x_scale();
            mtow(1) *= scalable->get_y_scale();
            mtow(2) *= scalable->get_z_scale();
        }

        proxy->set_transformation(mtow);
    }


    void set_transformation_stub2(
        ObjectEntityProxy *proxy,
        const ScalableObject *object)
    {
        set_transformation_stub1(
            proxy, dynamic_cast<const SceneObject *>(object));
    }


    void set_opacity_stub(
        ObjectEntityProxy *proxy,
        const SceneObject *object)
    {
        proxy->set_opacity(object->get_accumulated_opacity());
    }
}






RendererRelay::Impl::Impl(const SceneObject *root)
    : m_renderer(new Renderer)
{
    add_subtree(root);
}


RendererRelay::Impl::~Impl()
{
    // TODO: Clean up! We're leaking memory as it is now.

    if( m_renderer )
    {
        delete m_renderer;
        m_renderer = 0;
    }
}


namespace
{
    renderer::api::ObjectEntityProxy *_proxy = 0;
}


//
// ----------------- Begin visitor methods ----------------
//



ObjectEntityProxy *RendererRelay::Impl::add_entity(const SceneObject *subject)
{
    _proxy = 0;
    const_cast<SceneObject *>(subject)->accept(this);
    return _proxy;
}



void RendererRelay::Impl::visit(CameraObject *object)
{
    CameraProxy *proxy = new CameraProxy(*m_renderer);
    _proxy = proxy;


    // Relay type-specific attributes
    proxy->set_fov(object->get_fov());
    proxy->set_orthographic(object->is_orthographic());
    proxy->set_zoom_distance(object->get_zoom_distance());
    proxy->set_near_plane(object->get_near_plane());
    proxy->set_far_plane(object->get_far_plane());


    // Set up relay connections
    record_connection(
        object,
        object->m_on_view_mode_set.connect(
            boost::bind(&CameraProxy::set_orthographic, proxy, _2)));

    record_connection(
        object,
        object->m_on_fov_set.connect(
            boost::bind(&CameraProxy::set_fov, proxy, _2)));

    record_connection(
        object,
        object->m_on_zoom_distance_set.connect(
            boost::bind(&CameraProxy::set_zoom_distance, proxy, _2)));

    record_connection(
        object,
        object->m_on_near_plane_set.connect(
            boost::bind(&CameraProxy::set_near_plane, proxy, _2)));

    record_connection(
        object,
        object->m_on_far_plane_set.connect(
            boost::bind(&CameraProxy::set_far_plane, proxy, _2)));
}

void RendererRelay::Impl::visit(ModelObject *object)
{
    _proxy = new ModelProxy(*m_renderer, object->get_expanded_model_filename());
}

void RendererRelay::Impl::visit(SphereObject *object)
{
    _proxy = new SphereProxy(*m_renderer);
}

void RendererRelay::Impl::visit(GroupObject *object)
{
    _proxy = new DummyProxy(*m_renderer);
}

void RendererRelay::Impl::visit(CubeObject *object)
{
    _proxy = new CubeProxy(*m_renderer);
}

void RendererRelay::Impl::visit(PolygonObject *object)
{
    _proxy = new PolygonProxy(*m_renderer);
}

void RendererRelay::Impl::visit(CylinderObject *object)
{
    _proxy = new CylinderProxy(*m_renderer);
}

void RendererRelay::Impl::visit(CircleObject *object)
{
    _proxy = new CircleProxy(*m_renderer);
}

void RendererRelay::Impl::visit(PolyLineObject *object)
{
    assert( false );
}

void RendererRelay::Impl::visit(PointSetObject *object)
{
    assert( false );
}

void RendererRelay::Impl::visit(Hinge *object)
{
    _proxy = new DummyProxy(*m_renderer);
}

void RendererRelay::Impl::visit(Slider *object)
{
    _proxy = new DummyProxy(*m_renderer);
}

void RendererRelay::Impl::visit(GridObject *object)
{
    LineMeshProxy *proxy = new LineMeshProxy(*m_renderer);
    _proxy = proxy;

    record_connection(
        object,
        object->m_on_segment_count_set.connect(
            boost::bind(&set_grid_vertices_stub, proxy, _1)));

    record_connection(
        object,
        object->m_on_grid_type_set.connect(
            boost::bind(&set_grid_vertices_stub, proxy, _1)));

    record_connection(
        object,
        object->m_on_central_angle_set.connect(
            boost::bind(&set_grid_vertices_stub, proxy, _1)));

    set_grid_vertices_stub(proxy, object);
}

void RendererRelay::Impl::visit(PointCloud *object)
{
    _proxy = new PointMeshProxy(*m_renderer);
}

void RendererRelay::Impl::visit(LineCloud *object)
{
    _proxy = new LineMeshProxy(*m_renderer);
}

void RendererRelay::Impl::visit(OccupancyGrid2D *object)
{
    OccupancyGrid2DProxy *proxy = new OccupancyGrid2DProxy(
        *m_renderer,
        object->get_cell_size(),
        object->get_free_color(),
        object->get_occupied_color());
    _proxy = proxy;

    //proxy->set_cells(object->get_cells());

    record_connection(
        object,
        object->m_on_cells_set.connect(
            boost::bind(&OccupancyGrid2DProxy::set_cells, proxy, _2)));
}

void RendererRelay::Impl::visit(Label *object)
{
    LabelProxy *proxy = new LabelProxy(*m_renderer);
    _proxy = proxy;

    proxy->set_text(object->get_text());
    proxy->set_alignment(object->get_alignment());

    record_connection(
        object,
        object->m_on_text_set.connect(
            boost::bind(&LabelProxy::set_text, proxy, _2)));

    record_connection(
        object,
        object->m_on_alignment_set.connect(
            boost::bind(&LabelProxy::set_alignment, proxy, _2)));
}



//
// ----------------- End of visitor methods ----------------
//



ObjectEntityProxy *RendererRelay::Impl::get_proxy(ObjectID id)
{
    ObjectToEntityMap::iterator it = m_object_to_entity_map.find(id);
    if( it == m_object_to_entity_map.end() )
        throw std::runtime_error(
            "No entity proxy for the requested object exists");
    else
        return it->second;
}


const ObjectEntityProxy *RendererRelay::Impl::get_proxy(ObjectID id) const
{
    ObjectToEntityMap::const_iterator it = m_object_to_entity_map.find(id);
    if( it == m_object_to_entity_map.end() )
        throw std::runtime_error(
            "No entity proxy for the requested object exists");
    else
        return it->second;
}


ObjectEntityProxy *RendererRelay::Impl::get_proxy(
    const SceneObject *object)
{
    return get_proxy(object->get_object_id());
}


const ObjectEntityProxy *RendererRelay::Impl::get_proxy(
    const SceneObject *object) const
{
    return get_proxy(object->get_object_id());
}


bool RendererRelay::Impl::has_proxy(ObjectID id) const throw()
{
    ObjectToEntityMap::const_iterator it = m_object_to_entity_map.find(id);
    return it != m_object_to_entity_map.end();
}


bool RendererRelay::Impl::has_proxy(const SceneObject *object) const throw()
{
    return has_proxy(object->get_object_id());
}


void RendererRelay::Impl::render(
    unsigned int w, unsigned int h, ObjectID camera_id,
    bool *visible_layers, renderer::FrameCollector *collector)
{
    m_renderer->render(
        w, h,
        get_proxy(camera_id)->entity_id(),
        visible_layers,
        collector);
}

uint32_t RendererRelay::Impl::pick(
    unsigned int w, unsigned int h, ObjectID camera_id,
    bool *visible_layers,
    int x, int y, float sensitivity)
{
    return m_renderer->pick(
        w, h,
        get_proxy(camera_id)->entity_id(),
        visible_layers,
        x, y, sensitivity);
}


void RendererRelay::Impl::initialize_renderer()
{
    m_renderer->initialize();
}


bool RendererRelay::Impl::has_pending_renderer_updates() const throw()
{
    return m_renderer->has_pending_updates();
}


void RendererRelay::Impl::process_renderer_updates()
{
    m_renderer->process_updates();
}


void RendererRelay::Impl::add_subtree(const SceneObject *subtree)
{
    ObjectEntityProxy *proxy = add_entity(subtree);

    if( proxy )
    {
        // Add SceneObject -> ObjectEntityProxy mapping
        m_object_to_entity_map[subtree->get_object_id()] = proxy;
    }

    //
    // Set up connections to forward data when it changes
    //
    Connections &connections = m_connection_map[subtree];

    connections.push_back(
        const_cast<SceneObject *>(subtree)->m_on_child_attached.connect(
            boost::bind(&RendererRelay::Impl::add_subtree, this, _1)));

    connections.push_back(
        const_cast<SceneObject *>(subtree)->m_on_detached.connect(
            boost::bind(&RendererRelay::Impl::remove_subtree, this, _1)));


    if( proxy )
    {
        if( subtree->get_parent() && has_proxy(subtree->get_parent()) )
            proxy->set_parent(*get_proxy(subtree->get_parent()));

        set_transformation_stub1(proxy, subtree);
        proxy->set_opacity(subtree->get_accumulated_opacity());
        proxy->set_color(subtree->get_color());
        proxy->set_layer(subtree->get_layer());
        proxy->set_name(subtree->get_object_id());
        proxy->set_visibility(subtree->is_visible());
        proxy->set_selected(subtree->is_selected());
        proxy->set_has_selected_ancestor(subtree->has_selected_ancestor());


        connections.push_back(
            const_cast<SceneObject *>(subtree)->m_on_mtow_changed.connect(
                boost::bind(&set_transformation_stub1, proxy, _1)));

        connections.push_back(
            const_cast<SceneObject *>(subtree)->m_on_accum_opacity_changed.connect(
                boost::bind(&set_opacity_stub, proxy, _1)));

        connections.push_back(
            const_cast<SceneObject *>(subtree)->m_on_color_set.connect(
                boost::bind(&ObjectEntityProxy::set_color, proxy, _2)));

        connections.push_back(
            const_cast<SceneObject *>(subtree)->m_on_layer_set.connect(
                boost::bind(&ObjectEntityProxy::set_layer, proxy, _2)));

        connections.push_back(
            const_cast<SceneObject *>(subtree)->m_on_visibility_changed.connect(
                boost::bind(&ObjectEntityProxy::set_visibility, proxy, _2)));

        connections.push_back(
            const_cast<SceneObject *>(subtree)->m_on_selected.connect(
                boost::bind(&ObjectEntityProxy::set_selected, proxy, _2)));

        connections.push_back(
            const_cast<SceneObject *>(
                subtree)->m_on_has_selected_ancestor_changed.connect(
                boost::bind(
                    &ObjectEntityProxy::set_has_selected_ancestor, proxy, _2)));


        const ScalableObject *scalable(
            dynamic_cast<const ScalableObject *>(subtree));
        if( scalable )
        {
            connections.push_back(
                const_cast<ScalableObject *>(scalable)->m_on_scale_set.connect(
                    boost::bind(&set_transformation_stub2, proxy, _1)));
        }


        const VertexObject *vertex_based(
            dynamic_cast<const VertexObject *>(subtree));
        if( vertex_based )
        {
            VertexBasedProxy *vp = dynamic_cast<VertexBasedProxy *>(proxy);
            assert( vp );

            vp->set_vertices(vertex_based->get_vertices());

            connections.push_back(
                const_cast<VertexObject *>(vertex_based)->m_on_vertices_set.connect(
                    boost::bind(&VertexBasedProxy::set_vertices, vp, _2)));
        }

        const LineBased *line_based = dynamic_cast<const LineBased *>(subtree);
        if( line_based )
        {
            LineBasedProxy *lp = dynamic_cast<LineBasedProxy *>(proxy);
            assert( lp );

            lp->set_line_style(
                line_based->get_line_width(),
                line_based->get_line_style().first,
                line_based->get_line_style().second);

            connections.push_back(
                const_cast<LineBased *>(line_based)->m_on_line_params_set.connect(
                    boost::bind(&LineBasedProxy::set_line_style, lp, _2, _3, _4)));
        }
    }


    const SceneObject::ChildContainer &children = subtree->get_children();
    for( SceneObject::ChildContainer::const_iterator it = children.begin();
         it != children.end(); ++it )
    {
        add_subtree(*it);
    }
}


void RendererRelay::Impl::remove_subtree(const SceneObject *subtree)
{
    if( has_proxy(subtree) )
    {
        // Disconnect all slots for object
        ConnectionMap::iterator conn_it = m_connection_map.find(subtree);
        assert( conn_it != m_connection_map.end() );
        Connections &connections = conn_it->second;

        for( Connections::iterator it = connections.begin();
             it != connections.end(); ++it )
        {
            it->disconnect();
        }

        m_connection_map.erase(conn_it);

        ObjectEntityProxy *proxy = get_proxy(subtree);

        // Erase SceneObject -> ObjectEntityProxy mapping
        m_object_to_entity_map.erase(subtree->get_object_id());

        // Delete ObjectEntityProxy (which will delete the proxied object as well)
        delete proxy;
    }


    const SceneObject::ChildContainer &children = subtree->get_children();
    for( SceneObject::ChildContainer::const_iterator it = children.begin();
         it != children.end(); ++it )
    {
        remove_subtree(*it);
    }
}



void RendererRelay::Impl::record_connection(
    const SceneObject *object,
    boost::signalslib::connection connection)
{
    m_connection_map[object].push_back(connection);
}


renderer::api::Renderer &RendererRelay::Impl::get_renderer()
{
    if( !m_renderer )
        throw std::runtime_error("Renderer not initialized");
    return *m_renderer;
}


const renderer::api::Renderer &RendererRelay::Impl::get_renderer() const
{
    if( !m_renderer )
        throw std::runtime_error("Renderer not initialized");
    return *m_renderer;
}
