/*
 * Copyright Staffan Gimåker 2006-2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef PEEKABOT_RENDERER_RELAY_IMPL_HH_INCLUDED
#define PEEKABOT_RENDERER_RELAY_IMPL_HH_INCLUDED


#include <map>
#include <vector>
#include <boost/utility.hpp>
#include <boost/thread_safe_signal.hpp>

#include "Types.hh"
#include "RendererRelay.hh"
#include "ObjectVisitor.hh"


namespace peekabot
{
    class SceneObject;

    namespace renderer
    {
        class FrameCollector;

        namespace api
        {
            class Renderer;
            class ObjectEntityProxy;
        }
    }



    /**
     * \internal
     *
     * \brief Implementation class for the RendererRelay class that takes care
     * of all the gory details of interfacing and forwarding changes to the
     * renderer.
     *
     * \sa RendererRelay
     */
    class RendererRelay::Impl : public ObjectVisitor,
                                public boost::noncopyable
    {
    public:
        Impl(const SceneObject *root);

        virtual ~Impl();

        /// \copydoc RendererRelay::render
        void render(
            unsigned int w, unsigned int h, ObjectID camera_id,
            bool *visible_layers, renderer::FrameCollector *collector);

        /// \copydoc RendererRelay::pick
        uint32_t pick(
            unsigned int w, unsigned int h, ObjectID camera_id,
            bool *visible_layers,
            int x, int y, float sensitivity);

        /// \copydoc RendererRelay::initialize_renderer
        void initialize_renderer();

        /// \copydoc RendererRelay::has_pending_renderer_updates
        bool has_pending_renderer_updates() const throw();

        /// \copydoc RendererRelay::process_renderer_updates
        void process_renderer_updates();



        renderer::api::ObjectEntityProxy *get_proxy(ObjectID id);

        const renderer::api::ObjectEntityProxy *get_proxy(ObjectID id) const;

        renderer::api::ObjectEntityProxy *get_proxy(
            const SceneObject *object);

        const renderer::api::ObjectEntityProxy *get_proxy(
            const SceneObject *object) const;

        bool has_proxy(ObjectID id) const throw();

        bool has_proxy(const SceneObject *object) const throw();

        renderer::api::Renderer &get_renderer();

        const renderer::api::Renderer &get_renderer() const;



        /// \name Visitor methods
        /// @{

        virtual void visit(CameraObject *object);

        virtual void visit(ModelObject *object);

        virtual void visit(SphereObject *object);

        virtual void visit(GroupObject *object);

        virtual void visit(CubeObject *object);

        virtual void visit(PolygonObject *object);

        virtual void visit(CylinderObject *object);

        virtual void visit(CircleObject *object);

        virtual void visit(PolyLineObject *object);

        virtual void visit(PointSetObject *object);

        virtual void visit(Hinge *object);

        virtual void visit(Slider *object);

        virtual void visit(GridObject *object);

        virtual void visit(PointCloud *object);

        virtual void visit(LineCloud *object);

        virtual void visit(OccupancyGrid2D *object);

        virtual void visit(Label *object);

        /// @}

    private:
        renderer::api::ObjectEntityProxy *add_entity(const SceneObject *object);

        void add_subtree(const SceneObject *subtree);

        void remove_subtree(const SceneObject *subtree);

        void record_connection(
            const SceneObject *object,
            boost::signalslib::connection connection);

    private:
        renderer::api::Renderer *m_renderer;

        typedef std::map<
            ObjectID,
            renderer::api::ObjectEntityProxy *> ObjectToEntityMap;

        ObjectToEntityMap m_object_to_entity_map;


        typedef std::vector<boost::signalslib::connection> Connections;
        typedef std::map<const SceneObject *, Connections> ConnectionMap;

        ConnectionMap m_connection_map;
    };
}


#endif // PEEKABOT_RENDERER_RELAY_IMPL_HH_INCLUDED
