/*
 * Copyright Anders Boberg 2007.
 * Copyright Staffan Gimåker 2008-2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <boost/bind.hpp>

#include "Decorator.hh"
#include "UserInterfaceUtilities.hh"
#include "Manipulator.hh"
#include "../actions/SetSelected.hh"
#include "../RendererRelay.hh"
#include "../renderer/api/Renderer.hh"


using namespace peekabot;
using namespace peekabot::gui;


Decorator::Decorator(UserInterfaceUtilities &uiu)
    : m_uiu(uiu)
{
    // Connect slots...
    // ... for select/deselect
    m_uiu.on_object_selected().connect(
        boost::bind(&Decorator::selection_changed, this, _1, true));
    m_uiu.on_object_deselected().connect(
        boost::bind(&Decorator::selection_changed, this, _1, false));
    // ... tool changed
    m_uiu.get_manipulator()->get_mode_signal().connect(
        boost::bind(&Decorator::mode_changed, this, _1));
    // ... coordinate system changed
    m_uiu.get_manipulator()->get_system_signal().connect(
        boost::bind(&Decorator::coordinate_system_changed, this, _1));
    // ... pivot changed
    m_uiu.get_manipulator()->get_pivot_signal().connect(
        boost::bind(&Decorator::pivot_changed, this, _1));
    // ... axes changed
    m_uiu.get_manipulator()->get_axis_signal().connect(
        boost::bind(&Decorator::axes_changed, this, _1));
}

Decorator::~Decorator()
{
    // Disconnect slots...
    m_uiu.on_object_selected().disconnect(
        boost::bind(&Decorator::selection_changed, this, _1, true));
    m_uiu.on_object_deselected().disconnect(
        boost::bind(&Decorator::selection_changed, this, _1, false));

    if(m_uiu.get_manipulator() != 0)
    {
        m_uiu.get_manipulator()->get_mode_signal().disconnect(
            boost::bind(&Decorator::mode_changed, this, _1));
        m_uiu.get_manipulator()->get_system_signal().disconnect(
            boost::bind(&Decorator::coordinate_system_changed, this, _1));
        m_uiu.get_manipulator()->get_pivot_signal().disconnect(
            boost::bind(&Decorator::pivot_changed, this, _1));
        m_uiu.get_manipulator()->get_axis_signal().disconnect(
            boost::bind(&Decorator::axes_changed, this, _1));
    }
}

void Decorator::selection_changed(ObjectID id, bool selected) throw()
{
    m_uiu.post_action(new SetSelected(id, selected));
}

void Decorator::mode_changed(ManipulatorMode mode) throw()
{
    renderer::api::Renderer &r = TheRendererRelay::instance().get_renderer();

    if( mode == MOVE_MODE )
    {
        r.set_transformation_guides_mode(
            renderer::MOVE_TRANSFORMATION_GUIDES);
        r.set_transformation_guides_coord_sys(
            m_uiu.get_manipulator()->get_system());
        r.set_transformation_guides_axes(
            m_uiu.get_manipulator()->get_axes());
    }
    else if( mode == ROTATE_MODE )
    {
        r.set_transformation_guides_mode(
            renderer::ROTATE_TRANSFORMATION_GUIDES);
        r.set_transformation_guides_coord_sys(
            m_uiu.get_manipulator()->get_system());
        r.set_transformation_guides_axes(
            m_uiu.get_manipulator()->get_axes());
        r.set_transformation_guides_pivot(
            m_uiu.get_manipulator()->get_pivot());
    }
    else
    {
        r.set_transformation_guides_mode(
            renderer::NO_TRANSFORMATION_GUIDES);
    }
}

void Decorator::coordinate_system_changed(CoordinateSystem coord_sys) throw()
{
    renderer::api::Renderer &r = TheRendererRelay::instance().get_renderer();
    r.set_transformation_guides_coord_sys(coord_sys);
}


void Decorator::pivot_changed(RotationCenterpoint pivot) throw()
{
    renderer::api::Renderer &r = TheRendererRelay::instance().get_renderer();
    r.set_transformation_guides_pivot(pivot);
}


void Decorator::axes_changed(Axis axes) throw()
{
    renderer::api::Renderer &r = TheRendererRelay::instance().get_renderer();
    r.set_transformation_guides_axes(axes);
}

