/*
 * Copyright Staffan Gimåker 2006-2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <sstream>
#include <GL/glew.h>

#include "RenderEngine.hh"
#include "RenderTraverser.hh"
#include "SceneAction.hh"
#include "entities/Camera.hh"
#include "../MessageHub.hh"
#include "../Configuration.hh"
#include "Statelet.hh"
#include "statelets/Texture2D.hh"
#include "statelets/Transformation.hh"
#include "FrameCollector.hh"


using namespace peekabot;
using namespace peekabot::renderer;


namespace
{
    const RGBColor DEFAULT_BACKGROUND_COLOR(1, 1, 1);
    const RGBColor DEFAULT_SELECTED_COLOR(0, 0.3, 1);
    const RGBColor DEFAULT_ANCESTOR_SELECTED_COLOR(0.5, 0.75, 1);
}



RenderEngine::RenderEngine()
    : m_scene_manager(m_scene)
{
    try
    {
        // Get color values
        m_config.m_bg_color = DEFAULT_BACKGROUND_COLOR;
        /*TheConfiguration::instance().get_option<RGBColor>(
          "gui.viewports.background_color", DEFAULT_BACKGROUND_COLOR);*/

        m_config.m_selected_color = DEFAULT_SELECTED_COLOR;
        /*TheConfiguration::instance().get_option<RGBColor>(
          "gui.viewports.selected_color", DEFAULT_SELECTED_COLOR);*/

        m_config.m_ancestor_selected_color = DEFAULT_ANCESTOR_SELECTED_COLOR;
        /*TheConfiguration::instance().get_option<RGBColor>(
          "gui.viewports.ancestor_selected_color", DEFAULT_ANCESTOR_SELECTED_COLOR);*/

        m_config.m_enable_multisampling =
            TheConfiguration::instance().get<bool>(
                "renderer.enable_multisampling", true);
    }
    catch(std::runtime_error &e)
    {
        TheMessageHub::instance().publish(
            ERROR_MESSAGE,
            "Failed to get RenderEngine configuration: %s (%s:%u)", 
            e.what(), __FILE__, __LINE__);
    }
}


RenderEngine::~RenderEngine()
{
}


void RenderEngine::queue_action(SceneAction *action) throw()
{
    boost::recursive_mutex::scoped_lock lock(m_action_queue_mutex);
    m_action_queue.push(action);
}


size_t RenderEngine::pending_action_count() const throw()
{
    boost::recursive_mutex::scoped_lock lock(m_action_queue_mutex);
    return m_action_queue.size();
}


void RenderEngine::process_actions() throw()
{
    m_on_process_actions(this);

    {
        boost::recursive_mutex::scoped_lock lock1(m_action_queue_mutex);
        // Lock render/pick calls out until we're done
        boost::recursive_mutex::scoped_lock lock2(m_mutex);

        SceneAction::Context context(m_scene, m_scene_manager);

        while( !m_action_queue.empty() )
        {
            SceneAction *action = m_action_queue.front();
            m_action_queue.pop();

            try
            {
                action->execute(context);
            }
            catch(std::exception &e)
            {
                TheMessageHub::instance().publish(
                    ERROR_MESSAGE,
                    "SceneAction threw an exception with message '" +
                    std::string(e.what()) + "' during execution (%s:%u)",
                    __FILE__, __LINE__);
            }
            catch(...)
            {
                TheMessageHub::instance().publish(
                    ERROR_MESSAGE,
                    "SceneAction threw an unknown exception during execution (%s:%u)",
                    __FILE__, __LINE__);
            }

            delete action;
        }
    }

    m_on_actions_processed(this);
}


void RenderEngine::init()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    TheMessageHub::instance().publish(
        DEBUG_MESSAGE, 
        "Initializing GLEW...");
   
    // Setting this magical, undocumented flag will enable GLEW to report
    // that OpenGL >1.2 exists on Macs with crippled hardware (e.g. the
    // Mac Mini, which only supports OpenGL 1.2 hardware accelerated).
    //
    // Note that the extra set of GL functions it reports are OK after setting
    // this flag are software rendered through their LLVM stack!
    // But software rendering is better than nothing I guess?
    glewExperimental = GL_TRUE;

    // Initialize GLEW (Note that it is IMPORTANT that we have a valid
    // OpenGL context before trying to initialize GLEW).
    GLenum err = glewInit();
    if( GLEW_OK != err )
    {
        // Problem: glewInit failed, something is seriously wrong.
        throw std::runtime_error((const char *)glewGetErrorString(err));
    }


    // Log GL version, vendor, renderer and extensions
    TheMessageHub::instance().publish(
        LOG_MESSAGE, 
        "OpenGL vendor: %s", glGetString(GL_VENDOR));

    TheMessageHub::instance().publish(
        LOG_MESSAGE, 
        "OpenGL renderer: %s", glGetString(GL_RENDERER));

    TheMessageHub::instance().publish(
        LOG_MESSAGE, 
        "OpenGL version: %s", glGetString(GL_VERSION));

    TheMessageHub::instance().publish(
        LOG_MESSAGE, 
        "OpenGL extensions: %s", glGetString(GL_EXTENSIONS));


    // We depend on glBlendFunc and glBlendColor - they're not available
    // until OpenGL 1.4, or through the ARB_imaging extension available in 
    // OpenGL 1.2. Make sure we have either one or abort!
    if( !(GLEW_VERSION_1_4 || (GLEW_ARB_imaging && GLEW_VERSION_1_2)) )
    {
        throw std::runtime_error(
            "OpenGL 1.4 or OpenGL 1.2 and the ARB_imaging "
            "extension are required to run peekabot");
    }

    // If we don't have VBO support, print a nice little informational
    // message saying so...
    if( !GLEW_ARB_vertex_buffer_object )
    {
        TheMessageHub::instance().publish(
            INFO_MESSAGE, 
            "VBO support not available");
    }

    init_gl_context();
}


void RenderEngine::init_gl_context()
{
    TheMessageHub::instance().publish(
        DEBUG_MESSAGE, 
        "Performing initial configuration of the OpenGL context...");

    // The OpenGL implementation (probably) isn't thread safe...
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    //
    // Dial in the right settings for a few states that aren't
    // handled by the statelet system (at the moment, at least)
    //
    glShadeModel(GL_SMOOTH);
    glClearStencil(0);
    glEnable(GL_LIGHT0);

    // We need this since we do scaling with matrices.
    glEnable(GL_NORMALIZE);

    Device::toggle_multisampling(true);//m_config.m_enable_multisampling);

    glEnable(GL_DEPTH_TEST);
    glEnable(GL_DITHER);
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
    //glHint(GL_POLYGON_SMOOTH_HINT, GL_FASTEST);

    glLightModelf(GL_LIGHT_MODEL_LOCAL_VIEWER, 1);
    glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR);



    State baseline;
    
    // Enable lighting by default
    baseline.set(new statelets::Lighting(true));
    // Disable two-sided lighting by default
    baseline.set(new statelets::LightModel(false));
    // Disable blending by default
    baseline.set(new statelets::BlendMode(GL_ONE, GL_ZERO));
    // Enable backface culling by default
    baseline.set(new statelets::Culling(true));
    // Enable depth testing by default, using the GL_LESS as the depth func
    baseline.set(new statelets::DepthTest(statelets::DepthTest::LESS));
    // Enable depth mask writing by default
    baseline.set(new statelets::DepthMask(true));
    // Enable diffuse reflectance material color tracking
    baseline.set(new statelets::ColorMaterial(true));
    baseline.set(new statelets::PointParams(3, true));
    baseline.set(new statelets::LineParams(1, false, 0, 0));
    baseline.set(new statelets::PolygonMode(
                     statelets::PolygonMode::FRONT_AND_BACK,
                     statelets::PolygonMode::FILL));
    baseline.set(new statelets::StencilTest(false));
    baseline.set(new statelets::StencilFunc(
                     GL_ALWAYS, 1, static_cast<GLuint>(-1),
                     GL_KEEP, GL_KEEP, GL_REPLACE));
    // No texture by default
    baseline.set(new statelets::Texture2D());

    // Default to disabling all vertex buffers
    baseline.set(
        new statelets::VertexPointer(boost::shared_ptr<VertexBuffer>()));
    baseline.set(
        new statelets::NormalPointer(boost::shared_ptr<VertexBuffer>()));
    baseline.set(
        new statelets::ColorPointer(boost::shared_ptr<VertexBuffer>()));
    baseline.set(
        new statelets::TexCoordPointer(boost::shared_ptr<VertexBuffer>()));

    // Default material:
    baseline.set(
        new statelets::Material(
            RGBAColor(0.8,0.8,0.8,1),
            RGBAColor(0.2,0.2,0.2,1),
            RGBAColor(0,0,0,1),
            20));

    State::set_baseline(baseline);
}


void RenderEngine::render(
    unsigned int w, unsigned int h,
    EntityID camera_id, 
    bool *visible_layers,
    FrameCollector *collector)
{
    if( w < 1 || h < 1 )
        return;
        
    assert( visible_layers != 0 );
    
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    Config config(m_config);


    process_actions();

    Camera &camera = get_camera(camera_id);
    // Setup viewport
    camera.set_viewport(0, 0, w, h);


    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);


    // Clear color, depth and stencil buffers
    glClearColor(config.m_bg_color.r, 
                 config.m_bg_color.g, 
                 config.m_bg_color.b, 0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

    // Set up the viewing transform and projection matrix
    use_camera(camera);


    // We're not doing picking, ignore object names
    State::OverrideSet os(State::get_override_set());
    os.insert(statelets::Name::get_type());
    State::push_baseline();
    State::set_baseline(State::get_baseline_state(), os);
    // Unconditionally reapply the baseline state
    State::apply_baseline();

    RenderTraverser rt(camera, visible_layers);
    m_scene_manager.traverse(camera, rt);

    rt.render_opaque();
    rt.render_translucent();
    rt.render_outlines(
        config.m_selected_color, config.m_ancestor_selected_color);
    rt.render_tripods();

    State::pop_baseline();

    if( collector )
        collect_frame(collector, w, h);
}


uint32_t RenderEngine::pick(
    unsigned int w, unsigned int h,
    EntityID camera_id, 
    bool *visible_layers, 
    int x, int y, float sensitivity)
{
    if( w < 1 || h < 1 )
        return 0;
        
    assert( visible_layers != 0 );

    boost::recursive_mutex::scoped_lock lock(m_mutex);

    Config config(m_config);


    process_actions();

    Camera &camera = get_camera(camera_id);
    // Setup viewport
    camera.set_viewport(0, 0, w, h);

  
    glMatrixMode(GL_PROJECTION);
    GLint viewport[4];
    glGetIntegerv(GL_VIEWPORT, viewport);
    glLoadIdentity();
    // OpenGL assumes that the origin is in the lower left 
    // corner of the screen, FLTK does not...
    gluPickMatrix(x, h-y-1, sensitivity, sensitivity, viewport);
    glMatrixMode(GL_MODELVIEW);


    GLuint hit_records[512];
    GLint hits;

    glSelectBuffer(sizeof(hit_records), hit_records);
    glRenderMode(GL_SELECT);
    glInitNames();
    glPushName(0);

    //
    // Render objects in the view frustum
    //

    // Set up the viewing transform and projection matrix
    use_camera(camera);


    // We don't care about colors, materials, textures, ...
    State::OverrideSet os(State::get_override_set());
    os.insert(statelets::Texture2D::get_type());
    os.insert(statelets::Material::get_type());
    os.insert(statelets::ColorMaterial::get_type());
    os.insert(statelets::Color::get_type());
    os.insert(statelets::Alpha::get_type());
    os.insert(statelets::BlendMode::get_type());
    os.insert(statelets::ColorPointer::get_type());
    os.insert(statelets::TexCoordPointer::get_type());
    os.insert(statelets::NormalPointer::get_type());
    os.insert(statelets::Lighting::get_type());
    os.insert(statelets::StencilTest::get_type());
    os.insert(statelets::StencilFunc::get_type());
    State::push_baseline();
    State::set_baseline(State::get_baseline_state(), os);
    // Unconditionally reapply the baseline state
    State::apply_baseline();


    RenderTraverser rt(camera, visible_layers);
    m_scene_manager.traverse(camera, rt);

    rt.render_opaque();
    rt.render_translucent();
    //rt.render_always_on_top();

    State::pop_baseline();


    //
    // Examine the results
    //
    hits = glRenderMode(GL_RENDER);


    // Extract the picked ID
    uint32_t picked_id = 0;
    GLuint picked_id_x = 0;
    for( int i = 0; i < hits; ++i )
    {
        GLuint name = hit_records[4*i + 3];
        if( name != 0 )
        {
            if( hit_records[4*i+1] < picked_id_x || picked_id == 0 )
            {
                picked_id = name;
                picked_id_x = hit_records[4*i+1];
            }
        }
    }


    // TODO
    // Clone camera or back up its settings.
    // Change to ortho mode, with a unit cube frustum scaled by sensitivity
    // ...
    return picked_id;
}


Camera &RenderEngine::get_camera(EntityID camera_id)
{
    if( !m_scene.has_entity(camera_id) )
    {
        std::stringstream ss;
        ss << "No such camera (EntityID " << camera_id << ")";
        throw std::runtime_error(ss.str());
    }

    Entity *entity = m_scene.get_entity(camera_id);
    
    Camera *camera = dynamic_cast<Camera *>(entity);
    if( camera == 0 )
        throw std::runtime_error("The given entity is not a camera");

    return *camera;
}


void RenderEngine::use_camera(Camera &camera) const throw()
{   
    // Load the projection matrix
    glMatrixMode(GL_PROJECTION);
    camera.apply_projection_matrix();

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    Matrix4f camera_matrix = Matrix4f(
            Vector4f( 0,  0, -1,  0),
            Vector4f(-1,  0,  0,  0),
            Vector4f( 0,  1,  0,  0),
            Vector4f( 0,  0,  0,  1)) * camera.get_viewer_mtow().inverse();
    glLoadMatrixf( 
        reinterpret_cast<const GLfloat *>(&camera_matrix(0,0)) ); // TODO: change to something that doesn't break when the unit we use changes

    // Set the light direction in *world* coordinates
    Vector4f pos(1, -1, 2, 0);
    glLightfv(GL_LIGHT0, GL_POSITION, &pos(0));

    Vector4f ambient(0.1, 0.1, 0.1, 1);
    glLightfv(GL_LIGHT0, GL_AMBIENT, &ambient(0));

    //Vector4r specular(0.5, 0.5, 0.5, 1);
    //glLightfv(GL_LIGHT0, GL_SPECULAR, &specular(0));
    //
    //Vector4r diffuse(0.9, 0.9, 0.9, 1);
    //glLightfv(GL_LIGHT0, GL_DIFFUSE, &diffuse(0));
}


void RenderEngine::collect_frame(
    FrameCollector *collector, unsigned int w, unsigned int h) const
{
    uint8_t pixbuf[w*h*3];
    
    glPixelStorei(GL_PACK_ALIGNMENT, 1);
    glReadBuffer(GL_BACK);
    
    if( collector->get_pixel_format() == FrameCollector::RGB )
        glReadPixels(0, 0, w, h, GL_RGB, GL_UNSIGNED_BYTE, pixbuf);
    else if( collector->get_pixel_format() == FrameCollector::BGR )
        glReadPixels(0, 0, w, h, GL_BGR, GL_UNSIGNED_BYTE, pixbuf);
    else
        TheMessageHub::instance().publish(
            WARNING_MESSAGE,
            "Unsupported FrameCollector pixel format used, %s (%s:%u)",
            "the buffer contents will be undefined", __FILE__, __LINE__);
    
    
    collector->collect_frame(w, h, pixbuf);
}
