/*
 * Copyright Staffan Gimåker 2006-2008.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef PEEKABOT_RENDERER_RENDER_ENGINE_HH_INCLUDED
#define PEEKABOT_RENDERER_RENDER_ENGINE_HH_INCLUDED


#include <queue>
#include <boost/thread/recursive_mutex.hpp>

#include "Types.hh"
#include "Scene.hh"
#include "SceneManager.hh"


namespace peekabot
{
    namespace renderer
    {
        class FrameCollector;
        class SceneAction;

        class RenderEngine
        {
        public:
            RenderEngine();

            ~RenderEngine();

            void queue_action(SceneAction *action) throw();

            void process_actions() throw();

            size_t pending_action_count() const throw();



            void init();

            void render(
                unsigned int w, unsigned int h, EntityID camera_id,
                bool *visible_layers, FrameCollector *collector);

            /*void render_offscreen(
                unsigned int w, unsigned int h, EntityID camera_id,
                bool *visible_layers, FrameCollector *collector);*/

            uint32_t pick(
                unsigned int w, unsigned int h, EntityID camera_id,
                bool *visible_layers,
                int x, int y, float sensitivity);


            /// \name Signals
            /// @{

            typedef boost::signal<void (const RenderEngine *)> OnProcessActions;
            mutable OnProcessActions m_on_process_actions;

            typedef boost::signal<void (const RenderEngine *)> OnActionsProcessed;
            mutable OnActionsProcessed m_on_actions_processed;

            /// @}

        private:
            struct Config
            {
                RGBColor m_bg_color;
                
                RGBColor m_selected_color;
                
                RGBColor m_ancestor_selected_color;
                
                bool m_enable_multisampling;
            };

        private:
            /**
             * \brief Perform initial state configuration on a pristine context.
             *
             * Sets up certain parameters that are the same for all objects - 
             * i.e. use smooth shading, use "local viewer" for lighting 
             * calculations.
             */
            void init_gl_context();

            Camera &get_camera(EntityID camera_id);

            void use_camera(Camera &camera) const throw();

            void collect_frame(
                FrameCollector *collector, 
                unsigned int w, unsigned int h) const;


        private:
            /**
             * \brief A queue of modifications to the scene, slanted for 
             * incorporation when process_actions() is called.
             *
             * Scene changes are queued by queue_action(SceneAction *) and
             * processed (and thus remvoed from this queue) by 
             * process_actions().
             */
            std::queue<SceneAction *> m_action_queue;

            /**
             * \brief Mutex that governs access to m_action_queue.
             *
             * Must be used to synchronize both read and write operations on 
             * m_action_queue.
             */
            mutable boost::recursive_mutex m_action_queue_mutex;

            /**
             * \brief Mutex governing access to everything except the 
             * action queue.
             */
            mutable boost::recursive_mutex m_mutex;

            Scene m_scene;

            SceneManager m_scene_manager;

            Config m_config;
        };
    }
}


#endif // PEEKABOT_RENDERER_RENDER_ENGINE_HH_INCLUDED
