/*
 * Copyright Staffan Gimåker 2008-2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Renderer.hh"
#include "../RenderEngine.hh"
#include "CameraProxy.hh"
#include "../SceneAction.hh"
#include "TransformationGuidesProxy.hh"


using namespace peekabot;
using namespace peekabot::renderer;
using namespace peekabot::renderer::api;


Renderer::Renderer()
    : m_engine(new RenderEngine),
      m_is_initialized(false),
      m_guides(new TransformationGuidesProxy(*this))
{
}


Renderer::~Renderer()
{
    // Order matters here: Must delete guides before engine
    if( m_guides )
    {
        delete m_guides;
        m_guides = 0;
    }

    if( m_engine )
    {
        delete m_engine;
        m_engine = 0;
    }
}


void Renderer::initialize()
{
    if( m_is_initialized )
        throw std::logic_error("The Renderer is already initialized");
    else
    {
        m_engine->init();
        m_is_initialized = true;
    }
}


void Renderer::render(
    unsigned int w, unsigned int h,
    EntityID camera_id,
    bool *visible_layers,
    FrameCollector *collector)
{
    if( !m_is_initialized )
        initialize();

    m_engine->render(w, h, camera_id, visible_layers, collector);
}


uint32_t Renderer::pick(
    unsigned int w, unsigned int h,
    EntityID camera_id,
    bool *visible_layers,
    int x, int y, float sensitivity)
{
    if( !m_is_initialized )
        initialize();

    return m_engine->pick(
        w, h, camera_id,
        visible_layers,
        x, y, sensitivity);
}


bool Renderer::has_pending_updates() const throw()
{
    return m_engine->pending_action_count() > 0;
}


void Renderer::process_updates()
{
    if( !m_is_initialized )
        initialize();

    m_engine->process_actions();
}


void Renderer::set_transformation_guides_mode(TransformationGuidesMode mode)
{
    m_guides->set_mode(mode);
}


void Renderer::set_transformation_guides_coord_sys(CoordinateSystem coord_sys)
{
    m_guides->set_coord_sys(coord_sys);
}


void Renderer::set_transformation_guides_pivot(RotationCenterpoint pivot)
{
    m_guides->set_pivot(pivot);
}


void Renderer::set_transformation_guides_axes(Axis axes)
{
    m_guides->set_axes(axes);
}


void Renderer::dispatch(SceneAction *action)
{
    m_engine->queue_action(action);
}
