/*
 * Copyright Staffan Gimåker 2007-2008.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <GL/glew.h>
#include <boost/scoped_array.hpp>

#include "../Mesh.hh"
#include "MeshEntities.hh"
#include "../PrepareRenderContext.hh"
#include "../../GeometryToolbox.hh"


using namespace peekabot;
using namespace peekabot::renderer;


MeshEntity::MeshEntity(boost::shared_ptr<Mesh> mesh) throw()
    : m_mesh(mesh)
{
}

MeshEntity::~MeshEntity()
{
}

void MeshEntity::get_renderables(PrepareRenderContext &context) const
{
    context.prepare(m_mesh.get());
}


void MeshEntity::set_vertices(const VertexBased::Vertices &vertices) throw()
{
    boost::shared_ptr<VertexBuffer> buf(new VertexBuffer);

    if( !vertices.empty() )
    {
        boost::scoped_array<float> data(new float[3*vertices.size()]);

        for( size_t i = 0; i < vertices.size(); ++i )
        {
            data[3*i+0] = vertices[i](0);
            data[3*i+1] = vertices[i](1);
            data[3*i+2] = vertices[i](2);
        }

        buf->set_data(
            data.get(),
            3*sizeof(float)*vertices.size(), 
            VertexBuffer::STATIC_DRAW);
    }

    m_mesh->set_vertices(buf);

    if( !m_mesh->has_indices() )
        m_mesh->set_element_count(vertices.size()/m_mesh->vertices_per_element());

    calculate_bvs(vertices);
}

void MeshEntity::add_vertices(const VertexBased::Vertices &vertices) throw()
{
    assert( false ); // not yet implemented!

    /*if( get_vertices.get_size() == 0 )
        set_vertices(vertices);
    else
    {
        boost::scoped_array<uint8_t> data(new uint8_t[get_vertices.get_size()]);
        buf.get_data(&data[0], m_size, 0);
        set_data(data, m_size, m_usage);
        delete[] data;
    }*/
}

void MeshEntity::get_vertices(VertexBased::Vertices &vertices) const throw()
{
    vertices.clear();

    if( !m_mesh->has_vertices() )
        return;

    const boost::shared_ptr<VertexBuffer> verts = m_mesh->get_vertices();
    const float *ptr = (const float *)verts->map();

    size_t N = m_mesh->get_element_count()*m_mesh->vertices_per_element();
    for( size_t i = 0; i < N; ++i )
        vertices.push_back( Vector3f(ptr[3*i+0], ptr[3*i+1], ptr[3*i+2]) );       

    verts->unmap();
}

void MeshEntity::calculate_bvs(const VertexBased::Vertices &vertices) throw()
{
    // TODO: doesn't neccesarily work for indexed primitives!

    //
    // Calculate the midpoint of the vertices
    //
    Vector3f midpoint;
    float r;

    geom::calculate_approximate_bounding_sphere(
        vertices.begin(), vertices.end(),
        midpoint, r);
    
    set_bounding_sphere(BoundingSphere(midpoint, r));
}

boost::shared_ptr<Mesh> MeshEntity::get_mesh() throw()
{
    return m_mesh;
}

const boost::shared_ptr<Mesh> 
MeshEntity::get_mesh() const throw()
{
    return m_mesh;
}


//


PointMeshEntity::PointMeshEntity(boost::shared_ptr<Mesh> mesh) throw()
    : MeshEntity(mesh)
{
}

PointMeshEntity::PointMeshEntity(const PointMeshEntity &x) throw()
    : MeshEntity(boost::shared_ptr<Mesh>(x.get_mesh()->clone()))
{
}

PointMeshEntity::~PointMeshEntity()
{
}

PointMeshEntity *PointMeshEntity::clone() const throw()
{
    return new PointMeshEntity(boost::shared_ptr<Mesh>(get_mesh()->clone()));
}

//

LineMeshEntity::LineMeshEntity(boost::shared_ptr<Mesh> mesh) throw()
    : MeshEntity(mesh)
{
}

LineMeshEntity::LineMeshEntity(const LineMeshEntity &x) throw()
    : MeshEntity(boost::shared_ptr<Mesh>(x.get_mesh()->clone()))
{
}

LineMeshEntity::~LineMeshEntity()
{
}

LineMeshEntity *LineMeshEntity::clone() const throw()
{
    return new LineMeshEntity(boost::shared_ptr<Mesh>(get_mesh()->clone()));
}


void LineMeshEntity::set_line_style(
    float line_width, 
    uint16_t stipple_pattern, 
    int stipple_factor)
{
    m_mesh->get_state().set(
        new statelets::LineParams(
            line_width, stipple_pattern != 0xFFFF,
            stipple_factor, stipple_pattern));
}


//

TriMeshEntity::TriMeshEntity(boost::shared_ptr<Mesh> mesh) throw()
    : MeshEntity(mesh)
{
}

TriMeshEntity::TriMeshEntity(const TriMeshEntity &x) throw()
    : MeshEntity(boost::shared_ptr<Mesh>(x.get_mesh()->clone()))
{
}

TriMeshEntity::~TriMeshEntity()
{
}

TriMeshEntity *TriMeshEntity::clone() const throw()
{
    return new TriMeshEntity(boost::shared_ptr<Mesh>(get_mesh()->clone()));
}

void TriMeshEntity::set_indices(const Indices &indices)
{
    boost::shared_ptr<IndexBuffer> buf(new IndexBuffer);

    buf->set_data(
        &indices[0], 
        sizeof(uint32_t)*indices.size(), 
        IndexBuffer::STATIC_DRAW);

    m_mesh->set_indices(buf);
    m_mesh->set_element_count(indices.size()/m_mesh->vertices_per_element());

    //calculate_bvs();
}

void TriMeshEntity::add_indices(const Indices &indices)
{
    assert( false );
}

void TriMeshEntity::get_indices(Indices &indices) const
{
    indices.clear();

    if( !m_mesh->has_indices() )
        return;

    const boost::shared_ptr<IndexBuffer> inds = m_mesh->get_indices();
    const uint32_t *ptr = (const uint32_t *)inds->map();

    size_t N = m_mesh->get_element_count();
    for( size_t i = 0; i < N; ++i )
        indices.push_back( ptr[i] );

    inds->unmap();
}
